#include "cctk.h"
#include "cctk_Parameters.h"
#include "cctk_Arguments.h"
#include "cctk_Functions.h"

module StarMapper_ADMSources

  implicit none

  CCTK_REAL,dimension(:),pointer :: pradius => Null()
  CCTK_REAL,dimension(:),pointer :: pradius_center => Null()
  CCTK_REAL,dimension(:),pointer :: pmass  => Null()
  CCTK_REAL,dimension(:),pointer :: prho   => Null()
  CCTK_REAL,dimension(:),pointer :: ppress => Null()
  CCTK_REAL,dimension(:),pointer :: ptemp => Null()
  CCTK_REAL,dimension(:),pointer :: peps   => Null()
  CCTK_REAL,dimension(:),pointer :: pvel   => Null()
  CCTK_REAL,dimension(:),pointer :: pye    => Null()
  CCTK_REAL,dimension(:),pointer :: pomega => Null()

  CCTK_REAL,dimension(:),allocatable :: psi_old

end module

subroutine CCCC_StarMapper_CaptureArrays(CCTK_ARGUMENTS)

  use StarMapper_ADMSources, &
   mod_pradius => pradius, &
   mod_pradius_center => pradius_center, &
   mod_pmass => pmass, &
   mod_prho => prho, &
   mod_ppress => ppress, &
   mod_ptemp => ptemp, &
   mod_peps => peps, &
   mod_pvel => pvel, &
   mod_pye => pye, &
   mod_pomega => pomega

  implicit none

  ! Have to be in before DECLARE_CCTK_ARGUMENTS to avoid gcc failures
  TARGET pradius
  TARGET pradius_center
  TARGET pmass
  TARGET prho
  TARGET ppress
  TARGET ptemp
  TARGET peps
  TARGET pvel
  TARGET pye
  TARGET pomega

  DECLARE_CCTK_ARGUMENTS

  ! capture pointers to global arrays for use by callback routines
  mod_pradius => pradius 
  mod_pradius_center => pradius_center 
  mod_pmass => pmass 
  mod_prho => prho 
  mod_ppress => ppress 
  mod_ptemp => ptemp 
  mod_peps => peps 
  mod_pvel => pvel 
  mod_pye => pye 
  mod_pomega => pomega

end subroutine

CCTK_INT function StarMapper_Set_Rho_ADM(cctkGH, size, source, x, y, z)

  use StarMapper_Module
  use StarMapper_ADMSources

  implicit none

  DECLARE_CCTK_PARAMETERS
  DECLARE_CCTK_FUNCTIONS

  CCTK_POINTER_TO_CONST, INTENT(IN) ::  cctkGH
  CCTK_INT             , INTENT(IN) ::  size
  CCTK_REAL            , INTENT(OUT), dimension(size) :: source
  CCTK_REAL            , INTENT(IN), dimension(size) ::  x
  CCTK_REAL            , INTENT(IN), dimension(size) ::  y
  CCTK_REAL            , INTENT(IN), dimension(size) ::  z

  integer :: zones,i

  CCTK_REAL :: radius, tempvel, cosf, sinf, sint, cost, rxy

  CCTK_REAL :: tenthalpy, twlorentz2, rho_ADM

  CCTK_REAL :: rotrad,rotomega,avrho
  CCTK_REAL :: rfe,drtrans,dropfac
  CCTK_REAL :: GRB_omega

  CCTK_REAL :: rho, press, eps, alp, gxx
  CCTK_REAL, dimension(3) :: vel

  zones = Num_Radial

  ! storage to remember last solution in Update_Sources
  ! allocate here since this is the only routine that will always be called
  allocate(psi_old(size))

  !$OMP PARALLEL DO DEFAULT(none)                                             &
  !$OMP SHARED(source,x,y,z,ppress,prho,peps,pradius_center,pvel,pradius,     &
  !$OMP        pomega, salp, smetric, sr, psi_old)                            &
  !$OMP PRIVATE(i,                                                            &
  !$OMP         radius, tempvel, cosf, sinf, sint, cost, rxy,                 &
  !$OMP         tenthalpy, twlorentz2, rho_ADM,                               &
  !$OMP         rotrad,rotomega,avrho,                                        &
  !$OMP         rfe,drtrans,dropfac,                                          &
  !$OMP         rho, press, eps, alp, gxx,                                    &
  !$OMP         vel)                                                          &
  !$OMP FIRSTPRIVATE(cctkGH, size, zones, initial_vel, Rotation_Type,         &
  !$OMP              RotLaw_Type, initial_rho_abs_min, starmapper_rho_min,    &
  !$OMP              szones, rotlaw_a,rotlaw_omega0,rotlaw_x,rotlaw_z,        &
  !$OMP              rotlaw_grb_rtrans,rotlaw_grb_dropfac,rotlaw_grb_drmatch)
  do i = 1,size
    radius = sqrt(x(i)**2 + y(i)**2 + z(i)**2)

! fluid scalars

    call StarMapper_map(rho,radius,prho,pradius_center,zones)
    call StarMapper_map(press,radius,ppress,pradius_center,zones)
    call StarMapper_map(eps,radius,peps,pradius_center,zones)


! fluid vectors

    call StarMapper_map(tempvel,radius,pvel,pradius,zones)

    rxy = sqrt( x(i)**2 + y(i)**2 )

    if(radius .le. 1.0d-5) then
       cosf = 0.0d0
       sinf = 0.0d0
       sint = 0.0d0
       cost = 0.0d0
    else
       if (rxy .le. 1.d-5) then
          cosf = 0.0d0
          sinf = 0.0d0
       else
          cosf = x(i) / rxy
          sinf = y(i) / rxy
       endif

       sint = rxy / radius

       if(abs(z(i)).lt.1.0d-5) then
          cost = 0.0d0
       else
          cost = z(i) / radius
       endif
    endif

    ! special treatment: set vel to zero in the center

    if (radius .gt. pradius(1)) then

       vel(1) = tempvel * cosf * sint
       vel(2) = tempvel * sinf * sint
       vel(3) = tempvel * cost
    else
       tempvel = pvel(1)/pradius(1) * radius

       vel(1) = tempvel * cosf * sint
       vel(2) = tempvel * sinf * sint
       vel(3) = tempvel * cost

       ! do not warn about this since TP commonly requires points close to the
       ! origin
       !write(*,"(1P10E15.6)") z(i),vel(3)
    endif

    if(initial_vel.eq.0) then
       vel(1) = 0.0d0
       vel(2) = 0.0d0
       vel(3) = 0.0d0
    endif

!rotation
    if(CCTK_EQUALS(Rotation_Type,"rotlaw")) then
       if(CCTK_EQUALS(RotLaw_Type,"cylindrical") .or. CCTK_EQUALS(RotLaw_Type,"shellular")) then

          if(CCTK_EQUALS(RotLaw_Type,"cylindrical")) then
             rotrad = rxy
          else if(CCTK_EQUALS(RotLaw_Type,"shellular")) then
             rotrad = radius
          else
             call CCTK_WARN(0,"Rotation Law type not implemented")
          endif

          rotomega = RotLaw_Omega0 * 1.0d0 / &
               (1.0d0 + ( rotrad/RotLaw_A )**2 )

          vel(1) = vel(1) - rxy*rotomega*sinf
          vel(2) = vel(2) + rxy*rotomega*cosf
       else if (CCTK_EQUALS(RotLaw_Type,"MM_D")) then

          rotrad = rxy

          rotomega = RotLaw_Omega0 * &
               RotLaw_X**2 / (rotrad**2 + RotLaw_X**2) * &
               RotLaw_Z**4 / (z(i)**4 + RotLaw_Z**4)

          vel(1) = vel(1) - rxy*rotomega*sinf
          vel(2) = vel(2) + rxy*rotomega*cosf

       endif



! map rotating progenitor omega
    else if(CCTK_EQUALS(Rotation_Type,"rotating")) then
       call StarMapper_map(rotomega,rxy,pomega,pradius_center,zones)
       vel(1) = vel(1) - rxy*rotomega*sinf
       vel(2) = vel(2) + rxy*rotomega*cosf

    else if(CCTK_EQUALS(Rotation_Type,"GRB")) then
       rfe = RotLaw_GRB_rtrans * length_factor
       dropfac = RotLaw_GRB_dropfac
       drtrans = RotLaw_GRB_drmatch * length_factor

       if(CCTK_EQUALS(RotLaw_Type,"cylindrical")) then
          rotrad = rxy
       else if(CCTK_EQUALS(RotLaw_Type,"shellular")) then
          rotrad = radius
       else
          call CCTK_WARN(0,"Rotation Law type not implemented")
       endif

       rotomega = GRB_omega(RotLaw_Omega0,RotLaw_A,rfe,drtrans,dropfac,&
            rotrad)

       vel(1) = vel(1) - rxy*rotomega*sinf
       vel(2) = vel(2) + rxy*rotomega*cosf


    endif

! atmosphere
    if(rho.le.initial_rho_abs_min &
       .or. rho.le.starmapper_rho_min) then
      vel(1) = 0.0d0
      vel(2) = 0.0d0
      vel(3) = 0.0d0
    endif


! metric data

    call StarMapper_map(alp,radius,salp,sr,szones)
    call StarMapper_map(gxx,radius,smetric,sr,szones)

    if (abs(gxx).gt.2.0d0) then
       call CCTK_WARN(0,"blah")
    endif

! ADM energy density
    tenthalpy = rho*(1+eps)+press
    twlorentz2 = 1.d0/(1.d0 - gxx*(vel(1)**2+vel(2)**2+vel(3)**2))

    rho_ADM = tenthalpy*twlorentz2 - press

! gxx = psi***4, we want psi**8 * rho_ADM to get conformal energy density
    source(i) = gxx**2 * rho_ADM
    psi_old(i) = gxx**(1.d0/4.d0)

  end do
  !$OMP END PARALLEL DO

  StarMapper_Set_Rho_ADM = 1

end function

CCTK_INT function StarMapper_Set_Momentum_Source(cctkGH, dir, size, source, x, y, z)

  use StarMapper_Module
  use StarMapper_ADMSources

  implicit none

  DECLARE_CCTK_PARAMETERS
  DECLARE_CCTK_FUNCTIONS

  CCTK_POINTER_TO_CONST, INTENT(IN) ::  cctkGH
  CCTK_INT             , INTENT(IN) ::  dir
  CCTK_INT             , INTENT(IN) ::  size
  CCTK_REAL            , INTENT(OUT), dimension(size) :: source
  CCTK_REAL            , INTENT(IN), dimension(size) ::  x
  CCTK_REAL            , INTENT(IN), dimension(size) ::  y
  CCTK_REAL            , INTENT(IN), dimension(size) ::  z

  integer :: zones,i

  CCTK_REAL :: radius, tempvel, cosf, sinf, sint, cost, rxy

  CCTK_REAL :: tenthalpy, twlorentz2, Sj

  CCTK_REAL :: rotrad,rotomega,avrho
  CCTK_REAL :: rfe,drtrans,dropfac
  CCTK_REAL :: GRB_omega

  CCTK_REAL :: rho, press, eps, alp, gxx
  CCTK_REAL, dimension(3) :: vel

  zones = Num_Radial


  ! check input
  if(dir < 0 .or. dir > 2) then
    call CCTK_WARN(0, "dir must be between 0 and 2 incl.")
    ! NOTREACHED
  end if

  !$OMP PARALLEL DO DEFAULT(none)                                             &
  !$OMP SHARED(source,x,y,z,prho,ppress,peps,pradius_center,pvel,pradius,     &
  !$OMP        pomega,salp, smetric, sr)                                      &
  !$OMP PRIVATE(i,                                                            &
  !$OMP         radius, tempvel, cosf, sinf, sint, cost, rxy,                 &
  !$OMP         tenthalpy, twlorentz2, Sj,                                    &
  !$OMP         rotrad,rotomega,avrho,                                        &
  !$OMP         rfe,drtrans,dropfac,                                          &
  !$OMP         rho, press, eps, alp, gxx,                                    &
  !$OMP         vel)                                                          &
  !$OMP FIRSTPRIVATE(cctkGH, size, dir, zones, initial_vel, Rotation_Type,    &
  !$OMP              RotLaw_Type, initial_rho_abs_min, starmapper_rho_min,    &
  !$OMP              szones,rotlaw_a,rotlaw_omega0,rotlaw_x,rotlaw_z,         &
  !$OMP              rotlaw_grb_dropfac,rotlaw_grb_drmatch,rotlaw_grb_rtrans)
  do i = 1,size
    radius = sqrt(x(i)**2 + y(i)**2 + z(i)**2)

! fluid scalars

    call StarMapper_map(rho,radius,prho,pradius_center,zones)
    call StarMapper_map(press,radius,ppress,pradius_center,zones)
    call StarMapper_map(eps,radius,peps,pradius_center,zones)


! fluid vectors

    call StarMapper_map(tempvel,radius,pvel,pradius,zones)

    rxy = sqrt( x(i)**2 + y(i)**2 )

    if(radius .le. 1.0d-5) then
       cosf = 0.0d0
       sinf = 0.0d0
       sint = 0.0d0
       cost = 0.0d0
    else
       if (rxy .le. 1.d-5) then
          cosf = 0.0d0
          sinf = 0.0d0
       else
          cosf = x(i) / rxy
          sinf = y(i) / rxy
       endif

       sint = rxy / radius

       if(abs(z(i)).lt.1.0d-5) then
          cost = 0.0d0
       else
          cost = z(i) / radius
       endif
    endif

    ! special treatment: set vel to zero in the center

    if (radius .gt. pradius(1)) then

       vel(1) = tempvel * cosf * sint
       vel(2) = tempvel * sinf * sint
       vel(3) = tempvel * cost
    else
       tempvel = pvel(1)/pradius(1) * radius

       vel(1) = tempvel * cosf * sint
       vel(2) = tempvel * sinf * sint
       vel(3) = tempvel * cost

       ! do not warn about this since TP commonly requires points close to the
       ! origin
       !write(*,"(1P10E15.6)") z(i),vel(3)
    endif

    if(initial_vel.eq.0) then
       vel(1) = 0.0d0
       vel(2) = 0.0d0
       vel(3) = 0.0d0
    endif

!rotation
    if(CCTK_EQUALS(Rotation_Type,"rotlaw")) then
       if(CCTK_EQUALS(RotLaw_Type,"cylindrical") .or. CCTK_EQUALS(RotLaw_Type,"shellular")) then

          if(CCTK_EQUALS(RotLaw_Type,"cylindrical")) then
             rotrad = rxy
          else if(CCTK_EQUALS(RotLaw_Type,"shellular")) then
             rotrad = radius
          else
             call CCTK_WARN(0,"Rotation Law type not implemented")
          endif

          rotomega = RotLaw_Omega0 * 1.0d0 / &
               (1.0d0 + ( rotrad/RotLaw_A )**2 )

          vel(1) = vel(1) - rxy*rotomega*sinf
          vel(2) = vel(2) + rxy*rotomega*cosf
       else if (CCTK_EQUALS(RotLaw_Type,"MM_D")) then

          rotrad = rxy

          rotomega = RotLaw_Omega0 * &
               RotLaw_X**2 / (rotrad**2 + RotLaw_X**2) * &
               RotLaw_Z**4 / (z(i)**4 + RotLaw_Z**4)

          vel(1) = vel(1) - rxy*rotomega*sinf
          vel(2) = vel(2) + rxy*rotomega*cosf

       endif



! map rotating progenitor omega
    else if(CCTK_EQUALS(Rotation_Type,"rotating")) then
       call StarMapper_map(rotomega,rxy,pomega,pradius_center,zones)
       vel(1) = vel(1) - rxy*rotomega*sinf
       vel(2) = vel(2) + rxy*rotomega*cosf

    else if(CCTK_EQUALS(Rotation_Type,"GRB")) then
       rfe = RotLaw_GRB_rtrans * length_factor
       dropfac = RotLaw_GRB_dropfac
       drtrans = RotLaw_GRB_drmatch * length_factor

       if(CCTK_EQUALS(RotLaw_Type,"cylindrical")) then
          rotrad = rxy
       else if(CCTK_EQUALS(RotLaw_Type,"shellular")) then
          rotrad = radius
       else
          call CCTK_WARN(0,"Rotation Law type not implemented")
       endif

       rotomega = GRB_omega(RotLaw_Omega0,RotLaw_A,rfe,drtrans,dropfac,&
            rotrad)

       vel(1) = vel(1) - rxy*rotomega*sinf
       vel(2) = vel(2) + rxy*rotomega*cosf


    endif

! atmosphere
    if(rho.le.initial_rho_abs_min &
       .or. rho.le.starmapper_rho_min) then
      vel(1) = 0.0d0
      vel(2) = 0.0d0
      vel(3) = 0.0d0
    endif


! metric data

    call StarMapper_map(alp,radius,salp,sr,szones)
    call StarMapper_map(gxx,radius,smetric,sr,szones)

    if (abs(gxx).gt.2.0d0) then
       call CCTK_WARN(0,"blah")
    endif

! ADM current density (aka S_j)
    tenthalpy = rho*(1+eps)+press
    twlorentz2 = 1.d0/(1.d0 - gxx*(vel(1)**2+vel(2)**2+vel(3)**2))

    Sj= tenthalpy*twlorentz2 * gxx*vel(dir+1) ! C to Fortran index convention

! gxx = psi***4, we want psi**6 * S_j to get conformal current density
!                alternatively: we want psi**10 S^j
    source(i) = gxx**(6.d0/4.d0) * Sj

  end do
  !$OMP END PARALLEL DO

  StarMapper_Set_Momentum_Source = 1

end function

CCTK_INT function StarMapper_Set_Initial_Guess_for_u(cctkGH, size, source, x, y, z)

  use StarMapper_Module
  use StarMapper_ADMSources

  implicit none

  DECLARE_CCTK_PARAMETERS
  DECLARE_CCTK_FUNCTIONS

  CCTK_POINTER_TO_CONST, INTENT(IN) ::  cctkGH
  CCTK_INT             , INTENT(IN) ::  size
  CCTK_REAL            , INTENT(OUT), dimension(size) :: source
  CCTK_REAL            , INTENT(IN), dimension(size) ::  x
  CCTK_REAL            , INTENT(IN), dimension(size) ::  y
  CCTK_REAL            , INTENT(IN), dimension(size) ::  z

  integer :: i

  CCTK_REAL :: radius

  CCTK_REAL :: gxx

  !$OMP PARALLEL DO DEFAULT(none)                                             &
  !$OMP SHARED(source,x,y,z, smetric, sr)                                     &
  !$OMP PRIVATE(i,                                                            &
  !$OMP         radius, gxx)                                                  &
  !$OMP FIRSTPRIVATE(cctkGH, size, szones)
  do i = 1,size
    radius = sqrt(x(i)**2 + y(i)**2 + z(i)**2)

! metric data

    call StarMapper_map(gxx,radius,smetric,sr,szones)

    if (abs(gxx).gt.2.0d0) then
       call CCTK_WARN(0,"blah")
    endif

! gxx = psi***4, psi = 1 + u
    source(i) = gxx**(1.d0/4.d0) - 1.d0

  end do
  !$OMP END PARALLEL DO

  StarMapper_Set_Initial_Guess_for_u = 1

end function

CCTK_INT function StarMapper_Update_Sources(cctkGH, size, rhoADM, JxADM, &
                                            JyADM, JzADM, psi, x, y, z)

  use StarMapper_Module
  use StarMapper_ADMSources

  implicit none

  DECLARE_CCTK_PARAMETERS
  DECLARE_CCTK_FUNCTIONS

  CCTK_POINTER_TO_CONST, INTENT(IN) ::  cctkGH
  CCTK_INT             , INTENT(IN) ::  size
  CCTK_REAL            , INTENT(INOUT), dimension(size) :: rhoADM
  CCTK_REAL            , INTENT(INOUT), dimension(size) :: JxADM, JyADM, JzADM
  CCTK_REAL            , INTENT(IN), dimension(size) ::  psi
  CCTK_REAL            , INTENT(IN), dimension(size) ::  x
  CCTK_REAL            , INTENT(IN), dimension(size) ::  y
  CCTK_REAL            , INTENT(IN), dimension(size) ::  z


  integer :: zones,i

  CCTK_REAL :: radius, tempvel, cosf, sinf, sint, cost, rxy

  CCTK_REAL :: tenthalpy, twlorentz2, rho_ADM
  CCTK_REAL, dimension(3) :: Sj

  CCTK_REAL :: rotrad,rotomega,avrho
  CCTK_REAL :: rfe,drtrans,dropfac
  CCTK_REAL :: GRB_omega

  CCTK_REAL :: rho, press, eps
  CCTK_REAL, dimension(3) :: vel

  CCTK_REAL :: drelpsi, drelpsi_max

  zones = Num_Radial

  ! TP passes in the old sources in rhoADM, JiADM
  ! TP passes in the current solution in psi_new
  ! we are to update the sources 

  drelpsi_max = -1.d0 ! we only use positive ones afterwards

  !$OMP PARALLEL DO DEFAULT(none)                                             &
  !$OMP SHARED(rhoADM,JxADM,JyADM,JzADM,psi,x,y,z,prho,ppress,peps,           &
  !$OMP        pradius_center,pvel,pradius,                                   &
  !$OMP        pomega,salp, smetric, sr, psi_old)                             &
  !$OMP PRIVATE(i,                                                            &
  !$OMP         radius, tempvel, cosf, sinf, sint, cost, rxy, drelpsi,        &
  !$OMP         tenthalpy, twlorentz2, Sj, rho_ADM,                           &
  !$OMP         rotrad,rotomega,avrho,                                        &
  !$OMP         rfe,drtrans,dropfac,                                          &
  !$OMP         rho, press, eps,                                              &
  !$OMP         vel)                                                          &
  !$OMP FIRSTPRIVATE(cctkGH, size, zones, initial_vel, Rotation_Type,         &
  !$OMP              RotLaw_Type, initial_rho_abs_min, starmapper_rho_min,    &
  !$OMP              szones,rotlaw_a,rotlaw_omega0,rotlaw_x,rotlaw_z,         &
  !$OMP              rotlaw_grb_dropfac,rotlaw_grb_drmatch,rotlaw_grb_rtrans) &
  !$OMP REDUCTION(max : drelpsi_max)
  do i = 1,size
    radius = sqrt(x(i)**2 + y(i)**2 + z(i)**2)

! fluid scalars

    call StarMapper_map(rho,radius,prho,pradius_center,zones)
    call StarMapper_map(press,radius,ppress,pradius_center,zones)
    call StarMapper_map(eps,radius,peps,pradius_center,zones)


! fluid vectors

    call StarMapper_map(tempvel,radius,pvel,pradius,zones)

    rxy = sqrt( x(i)**2 + y(i)**2 )

    if(radius .le. 1.0d-5) then
       cosf = 0.0d0
       sinf = 0.0d0
       sint = 0.0d0
       cost = 0.0d0
    else
       if (rxy .le. 1.d-5) then
          cosf = 0.0d0
          sinf = 0.0d0
       else
          cosf = x(i) / rxy
          sinf = y(i) / rxy
       endif

       sint = rxy / radius

       if(abs(z(i)).lt.1.0d-5) then
          cost = 0.0d0
       else
          cost = z(i) / radius
       endif
    endif

    ! special treatment: set vel to zero in the center

    if (radius .gt. pradius(1)) then

       vel(1) = tempvel * cosf * sint
       vel(2) = tempvel * sinf * sint
       vel(3) = tempvel * cost
    else
       tempvel = pvel(1)/pradius(1) * radius

       vel(1) = tempvel * cosf * sint
       vel(2) = tempvel * sinf * sint
       vel(3) = tempvel * cost

       ! do not warn about this since TP commonly requires points close to the
       ! origin
       !write(*,"(1P10E15.6)") z(i),vel(3)
    endif

    if(initial_vel.eq.0) then
       vel(1) = 0.0d0
       vel(2) = 0.0d0
       vel(3) = 0.0d0
    endif

!rotation
    if(CCTK_EQUALS(Rotation_Type,"rotlaw")) then
       if(CCTK_EQUALS(RotLaw_Type,"cylindrical") .or. CCTK_EQUALS(RotLaw_Type,"shellular")) then

          if(CCTK_EQUALS(RotLaw_Type,"cylindrical")) then
             rotrad = rxy
          else if(CCTK_EQUALS(RotLaw_Type,"shellular")) then
             rotrad = radius
          else
             call CCTK_WARN(0,"Rotation Law type not implemented")
          endif

          rotomega = RotLaw_Omega0 * 1.0d0 / &
               (1.0d0 + ( rotrad/RotLaw_A )**2 )

          vel(1) = vel(1) - rxy*rotomega*sinf
          vel(2) = vel(2) + rxy*rotomega*cosf
       else if (CCTK_EQUALS(RotLaw_Type,"MM_D")) then

          rotrad = rxy

          rotomega = RotLaw_Omega0 * &
               RotLaw_X**2 / (rotrad**2 + RotLaw_X**2) * &
               RotLaw_Z**4 / (z(i)**4 + RotLaw_Z**4)

          vel(1) = vel(1) - rxy*rotomega*sinf
          vel(2) = vel(2) + rxy*rotomega*cosf

       endif



! map rotating progenitor omega
    else if(CCTK_EQUALS(Rotation_Type,"rotating")) then
       call StarMapper_map(rotomega,rxy,pomega,pradius_center,zones)
       vel(1) = vel(1) - rxy*rotomega*sinf
       vel(2) = vel(2) + rxy*rotomega*cosf

    else if(CCTK_EQUALS(Rotation_Type,"GRB")) then
       rfe = RotLaw_GRB_rtrans * length_factor
       dropfac = RotLaw_GRB_dropfac
       drtrans = RotLaw_GRB_drmatch * length_factor

       if(CCTK_EQUALS(RotLaw_Type,"cylindrical")) then
          rotrad = rxy
       else if(CCTK_EQUALS(RotLaw_Type,"shellular")) then
          rotrad = radius
       else
          call CCTK_WARN(0,"Rotation Law type not implemented")
       endif

       rotomega = GRB_omega(RotLaw_Omega0,RotLaw_A,rfe,drtrans,dropfac,&
            rotrad)

       vel(1) = vel(1) - rxy*rotomega*sinf
       vel(2) = vel(2) + rxy*rotomega*cosf


    endif

! atmosphere
    if(rho.le.initial_rho_abs_min &
       .or. rho.le.starmapper_rho_min) then
      vel(1) = 0.0d0
      vel(2) = 0.0d0
      vel(3) = 0.0d0
    endif


! shorthands
    tenthalpy = rho*(1+eps)+press
    twlorentz2 = 1.d0/(1.d0 - psi(i)**4*(vel(1)**2+vel(2)**2+vel(3)**2))

! ADM energy density
    rho_ADM = tenthalpy*twlorentz2 - press

! we want psi**8 * rho_ADM to get conformal energy density
    rhoADM(i) = psi(i)**8 * rho_ADM

! ADM current density (aka S_j)
    Sj = tenthalpy*twlorentz2 * psi(i)**4*vel

! we want psi**6 * S_j to get conformal current density
! alternatively: we want psi**10 S^j
    JxADM(i) = psi(i)**6 * Sj(1)
    JxADM(i) = psi(i)**6 * Sj(2)
    JxADM(i) = psi(i)**6 * Sj(3)

! gxx = psi***4
    drelpsi = abs(psi(i) - psi_old(i)) / psi_old(i)
    drelpsi_max = max(drelpsi_max, drelpsi)

    psi_old(i) = psi(i)

  end do
  !$OMP END PARALLEL DO

  ! ask TP to iterate again if the tolerance has not yet been reached
  if(drelpsi_max .le. metric_iteration_reltol) then
    StarMapper_Update_Sources = 0
  else
    StarMapper_Update_Sources = 1
  end if

end function
