#include "cctk.h"
#include "cctk_Parameters.h"
#include "cctk_Arguments.h"
#include "cctk_Functions.h"

subroutine CCCC_StarMapper_ConvertUnits(CCTK_ARGUMENTS)

  use StarMapper_Module

  implicit none

  DECLARE_CCTK_ARGUMENTS
  DECLARE_CCTK_PARAMETERS
  DECLARE_CCTK_FUNCTIONS

  integer i
  

  if (CCTK_EQUALS(Profile_Type,"SESAME") .or. &
       CCTK_EQUALS(Profile_Type,"PolyCott") .or. &
       CCTK_EQUALS(Profile_Type,"HegerShort")) then
     call ConvertUnitsSESAME(pradius,pmass,prho,&
          ppress,ptemp,peps,pvel,pye,pomega,num_radial)
  else
     call CCTK_WARN(0,"Profile Type not implemented!")
  end if

  pradius_center(1) = pradius(1)/2.0d0
  do i=2,Num_Radial
     pradius_center(i) = (pradius(i) + pradius(i-1))/2.0d0
  enddo

  ! needed for getting 2006 results
#if 0
  pradius_center = pradius
#endif 

end subroutine CCCC_StarMapper_ConvertUnits

subroutine ConvertUnitsSESAME(pradius,pmass,prho, &
          ppress,ptemp,peps,pvel,pye,pomega,zones)

  use StarMapper_Module

  implicit none

  integer zones
  integer i
  real*8 tye

  CCTK_REAL pradius(zones),pmass(zones),prho(zones),ppress(zones),ptemp(zones)
  CCTK_REAL peps(zones),pvel(zones),pye(zones),pomega(zones)

  pradius = length_factor * pradius
  pmass = mass_factor * pmass
  prho = density_factor * prho
  ppress = press_factor * ppress

! we are not converting ptemp; keeping it in mev
! ptemp
  
  peps = specific_internal_energy_factor * peps
  pvel = velocity_factor * pvel

! of course we are also not converting ye
! pye

  pomega = 1.0d0/time_factor * pomega

end subroutine ConvertUnitsSESAME
