#include "cctk.h"
#include "cctk_Parameters.h"
#include "cctk_Arguments.h"
#include "cctk_Functions.h"
#define CLIGHT 2.99792458e10

function GRB_lam(r,drtrans,rfe)
  implicit none
  CCTK_REAL GRB_lam, r, drtrans, rfe
  GRB_lam = (1.0 + tanh((r-rfe)/drtrans))/2.0
end function GRB_lam

function GRB_omega(omega0,A,rfe,drtrans,dropfac,r)
  
  implicit none
  CCTK_REAL GRB_omega,omega0,A,rfe,drtrans,dropfac,r
  CCTK_REAL GRB_lam

  GRB_omega = (1.0-GRB_lam(r,drtrans,rfe)) * (omega0 / (1 + (r/A)**(2))) &
              + GRB_lam(r,drtrans,rfe)*(omega0/((1 + (rfe/A)**(2)))/dropfac &
              /  (1 + (abs(r-rfe)/A)**(1.0/3.0)))


end function GRB_omega

subroutine CCCC_StarMapper_Map1D3D(CCTK_ARGUMENTS)

  use StarMapper_Module
  
  implicit none

  DECLARE_CCTK_ARGUMENTS
  DECLARE_CCTK_PARAMETERS
  DECLARE_CCTK_FUNCTIONS

  integer zones,i,j,k,l,nx,ny,nz
  integer upper_index,lower_index,istat

  CCTK_REAL radius, tempvel, cosf, sinf, sint, cost, rxy
  CCTK_REAL StarMapper_linterp

  CCTK_REAL tenthalpy, twlorentz, rdetg, tye

  CCTK_REAL rotrad,rotomega,avrho
  CCTK_REAL rfe,drtrans,dropfac
  CCTK_REAL GRB_omega

  logical :: set_metric, set_lapse, set_shift, set_hydro

  zones = Num_Radial

  nx = cctk_lsh(1)
  ny = cctk_lsh(2)
  nz = cctk_lsh(3)
                
  call CCTK_INFO("Mapping 1D profile to 3D")

! we are omitting Omega for the time being

  set_metric = set_everything .ne.0 .or. CCTK_EQUALS(initial_data, "StarMapper")
  set_lapse = set_everything .ne.0 .or. CCTK_EQUALS(initial_lapse, "StarMapper")
  set_shift = set_everything .ne.0 .or. CCTK_EQUALS(initial_shift, "StarMapper")
  set_hydro = set_everything .ne.0 .or. CCTK_EQUALS(initial_hydro, "StarMapper")


  if(set_metric) then
    gxx(:,:,:) = 0.0d0
    gxy(:,:,:) = 0.0d0
    gxz(:,:,:) = 0.0d0
    gyy(:,:,:) = 0.0d0
    gyz(:,:,:) = 0.0d0
    gzz(:,:,:) = 0.0d0

    kxx(:,:,:) = 0.0d0
    kxy(:,:,:) = 0.0d0
    kxz(:,:,:) = 0.0d0
    kyy(:,:,:) = 0.0d0
    kyz(:,:,:) = 0.0d0
    kzz(:,:,:) = 0.0d0
  end if

  if(set_shift) then
    betax(:,:,:) = 0.0d0
    betay(:,:,:) = 0.0d0
    betaz(:,:,:) = 0.0d0
  end if

#if 0
  radius = 15.0d0
  cost=1.0d0
  open(unit=667,file="blah.out")
  do i=1,50
     call StarMapper_map(tempvel,radius,pvel,pradius,zones)
     write(667,"(i5,1P10E15.6)") i,radius,tempvel
     radius = radius + cost
  enddo
  close(667)
#endif 

  do i=1,nx
     do j=1,ny
        do k=1,nz
           radius = sqrt(x(i,j,k)**2 + y(i,j,k)**2 + z(i,j,k)**2)

           if(set_hydro) then
  ! fluid scalars

             call StarMapper_map(rho(i,j,k),radius,prho,pradius_center,zones)

  ! check for rho_min


             call StarMapper_map(press(i,j,k),radius,ppress,pradius_center,zones)
             call StarMapper_map(eps(i,j,k),radius,peps,pradius_center,zones)

  ! Temperature
             if(Starmapper_eoskey.eq.4) then
                call StarMapper_map(temperature(i,j,k),radius,ptemp,pradius_center,zones)
             endif

  ! Y_e
             istat = 0
             call CCTK_QueryGroupStorage(istat,cctkGH,"hydrobase::Y_e")
             if(istat.ne.0) then
                call StarMapper_map(y_e(i,j,k),radius,pye,pradius_center,zones)
             endif


  ! fluid vectors
             
             call StarMapper_map(tempvel,radius,pvel,pradius,zones)

             rxy = sqrt( x(i,j,k)**2 + y(i,j,k)**2 ) 

             if(radius .le. 1.0d-5) then
                cosf = 0.0d0
                sinf = 0.0d0
                sint = 0.0d0
                cost = 0.0d0
             else
                if (rxy .le. 1.d-5) then
                   cosf = 0.0d0
                   sinf = 0.0d0
                else 
                   cosf = x(i,j,k) / rxy
                   sinf = y(i,j,k) / rxy
                endif

                sint = rxy / radius
                
                if(abs(z(i,j,k)).lt.1.0d-5) then
                   cost = 0.0d0
                else
                   cost = z(i,j,k) / radius                 
                endif
             endif

             ! special treatment: set vel to zero in the center

             if (radius .gt. pradius(1)) then

                vel(i,j,k,1) = tempvel * cosf * sint
                vel(i,j,k,2) = tempvel * sinf * sint
                vel(i,j,k,3) = tempvel * cost
             else 
                tempvel = pvel(1)/pradius(1) * radius

                vel(i,j,k,1) = tempvel * cosf * sint
                vel(i,j,k,2) = tempvel * sinf * sint
                vel(i,j,k,3) = tempvel * cost

  !              write(*,"(1P10E15.6)") z(i,j,k),velz(i,j,k)
             endif

             if(initial_vel.eq.0) then
                vel(i,j,k,1) = 0.0d0
                vel(i,j,k,2) = 0.0d0
                vel(i,j,k,3) = 0.0d0
             endif

  !rotation
             if(CCTK_EQUALS(Rotation_Type,"rotlaw")) then
                if(CCTK_EQUALS(RotLaw_Type,"cylindrical") .or. CCTK_EQUALS(RotLaw_Type,"shellular")) then

                   if(CCTK_EQUALS(RotLaw_Type,"cylindrical")) then
                      rotrad = rxy
                   else if(CCTK_EQUALS(RotLaw_Type,"shellular")) then
                      rotrad = radius
                   else 
                      call CCTK_WARN(0,"Rotation Law type not implemented")
                   endif
                   
                   rotomega = RotLaw_Omega0 * 1.0d0 / &
                        (1.0d0 + ( rotrad/RotLaw_A )**2 )

                   vel(i,j,k,1) = vel(i,j,k,1) - rxy*rotomega*sinf
                   vel(i,j,k,2) = vel(i,j,k,2) + rxy*rotomega*cosf
                else if (CCTK_EQUALS(RotLaw_Type,"MM_D")) then

                   rotrad = rxy

                   rotomega = RotLaw_Omega0 * &
                        RotLaw_X**2 / (rotrad**2 + RotLaw_X**2) * &
                        RotLaw_Z**4 / (z(i,j,k)**4 + RotLaw_Z**4)

                   vel(i,j,k,1) = vel(i,j,k,1) - rxy*rotomega*sinf
                   vel(i,j,k,2) = vel(i,j,k,2) + rxy*rotomega*cosf
                   
                endif
                   

                   
  !              write(*,"(1P10E15.6)") x(i,j,k), y(i,j,k), velx(i,j,k), &
  !                   vely(i,j,k),rxy*rotomega*sinf,rxy*rotomega*cosf

  ! map rotating progenitor omega
             else if(CCTK_EQUALS(Rotation_Type,"rotating")) then
                call StarMapper_map(rotomega,rxy,pomega,pradius_center,zones)
                vel(i,j,k,1) = vel(i,j,k,1) - rxy*rotomega*sinf
                vel(i,j,k,2) = vel(i,j,k,2) + rxy*rotomega*cosf

             else if(CCTK_EQUALS(Rotation_Type,"GRB")) then
                rfe = RotLaw_GRB_rtrans * length_factor
                dropfac = RotLaw_GRB_dropfac
                drtrans = RotLaw_GRB_drmatch * length_factor

                if(CCTK_EQUALS(RotLaw_Type,"cylindrical")) then
                   rotrad = rxy
                else if(CCTK_EQUALS(RotLaw_Type,"shellular")) then
                   rotrad = radius
                else 
                   call CCTK_WARN(0,"Rotation Law type not implemented")
                endif

                rotomega = GRB_omega(RotLaw_Omega0,RotLaw_A,rfe,drtrans,dropfac,&
                     rotrad)
                
                vel(i,j,k,1) = vel(i,j,k,1) - rxy*rotomega*sinf
                vel(i,j,k,2) = vel(i,j,k,2) + rxy*rotomega*cosf
                

             endif

  ! atmosphere
             if(rho(i,j,k).le.initial_rho_abs_min &
                  .or. rho(i,j,k).le.starmapper_rho_min) then
  !              rho(i,j,k) = initial_rho_abs_min
                vel(i,j,k,1) = 0.0d0
                vel(i,j,k,2) = 0.0d0
                vel(i,j,k,3) = 0.0d0
  !              tracer(i,j,k,1) = 0.5d0
             endif
          end if


! metric data
          if(set_lapse) then
            call StarMapper_map(alp(i,j,k),radius,salp,sr,szones)
          end if
          if(set_metric) then
            call StarMapper_map(gxx(i,j,k),radius,smetric,sr,szones)

!            write(*,*) gxx(i,j,k), i,j,k, radius

             if (abs(gxx(i,j,k)).gt.2.0d0) then
  !              do l=1,zones
  !                 write(*,*) l,pmetric(l)
  !              enddo
                call CCTK_WARN(0,"blah")
             endif

             gyy(i,j,k) = gxx(i,j,k)
             gzz(i,j,k) = gxx(i,j,k)
          end if

        enddo
     enddo
  enddo

end subroutine CCCC_StarMapper_Map1D3D

subroutine StarMapper_map(point_value,point_radius,parray,pradius,zones)

  implicit none

  CCTK_REAL point_value, point_radius
  CCTK_REAL pradius(*), parray(*)
  integer zones
  integer upper_index, lower_index

  if (point_radius .ge. pradius(1) .and. point_radius .lt. pradius(zones) ) then

     call StarMapper_find_index(zones,pradius,point_radius,upper_index,lower_index)

!     write(*,*) "indicies",upper_index,lower_index

     call StarMapper_linterp( pradius(lower_index),pradius(upper_index), &
          parray(lower_index), parray(upper_index), point_radius, point_value )

!     write(*,*) "point_value",point_value, point_radius

  else if (point_radius .lt. pradius(1)) then
! linear extrapolation

     call StarMapper_linterp(pradius(1),pradius(2), & 
          parray(1),parray(2),point_radius,point_value)

!     write(*,*) "point_value A",point_value, point_radius

  else if (point_radius .gt. pradius(zones)) then
! linear extrapolation
     call StarMapper_linterp(pradius(zones-1),pradius(zones), & 
          parray(zones-1),parray(zones),point_radius,point_value)

!     write(*,*) "point_value B",point_value, point_radius,pradius(zones)
  endif


end subroutine

subroutine StarMapper_linterp(x1,x2,y1,y2,x,y)

  implicit none

  CCTK_REAL slope,x1,x2,y1,y2,x,y

  if (x2.lt.x1) then
     call CCTK_WARN(0,"Error in linterp!")
  endif

  slope = (y2 - y1) / (x2 - x1)

!  write(*,*) "slope", slope, x1,x2,y1,y2,x

  y = slope*(x-x1) + y1
 
end subroutine  StarMapper_linterp


subroutine StarMapper_find_index(zones,array,goal,upper_index,lower_index)
  
  implicit none

  integer zones,i
  CCTK_REAL array(*)
  CCTK_REAL goal
  CCTK_INT middle_index,upper_index,lower_index

  lower_index = 1
  upper_index = zones

  do while ( (upper_index - lower_index) .gt. 1 )
     middle_index = (lower_index + upper_index) * 0.5
     if ( (goal .ge. array(lower_index)) &
          .and. (goal .le. array(middle_index)) ) then
        upper_index = middle_index
     else
        if ( (goal .ge. array(middle_index)) &
          .and. (goal .le. array(upper_index)) ) then
           lower_index = middle_index
        endif
     endif
  enddo

end subroutine StarMapper_find_index
