#include "cctk.h"
#include "cctk_Parameters.h"
#include "cctk_Arguments.h"
#include "cctk_Functions.h"

subroutine CCCC_StarMapper_NewtPot(CCTK_ARGUMENTS)

  use StarMapper_Module
  implicit none


  DECLARE_CCTK_ARGUMENTS
  DECLARE_CCTK_PARAMETERS
  DECLARE_CCTK_FUNCTIONS

  integer zones,i,ierr,maxi

  CCTK_REAL xmin,xmax,ymin,ymax,zmin,zmax,rmax,rmaxpot

  CCTK_REAL ggrav,pi,dr, difference, rhoav

  pi=3.14159265358979d0
  ggrav=1.0d0 !we are calculating this in Cactus units!

  maxi = 0

  call CCTK_INFO("Calculating Newtonian Potential")

  sr = 0.0d0
  srho = 0.0d0
  snewtpot = 0.0d0

  zones = Num_Radial


  call CCTK_COORDRANGE(ierr,cctkGH,xmin,xmax,1,"x","cart3d")
  call CCTK_COORDRANGE(ierr,cctkGH,ymin,ymax,2,"y","cart3d")
  call CCTK_COORDRANGE(ierr,cctkGH,zmin,zmax,3,"z","cart3d")
  

  rmaxpot = sqrt(xmax**2 + ymax**2 + zmax**2) + 50.0d0
  rmaxpot = 4000.0d0

  maxi=0
  i=1
  do while(maxi.eq.0)
     i=i+1
     if(pradius(i).ge.rmaxpot) then
        maxi=i
     endif
  enddo

  szones = Num_Radial+Num_Spacetime-maxi

  rmax = pradius(maxi-1)

  dr = rmax/(Num_Spacetime-1)

  sr(1) = 0.0d0
  do i=2,Num_Spacetime
     sr(i) = sr(i-1) + dr
  enddo

  do i=Num_Spacetime+1,szones
     sr(i) = pradius(i-Num_Spacetime+maxi)
     srho(i) = prho(i-Num_Spacetime+maxi)
  enddo

  
!  rhoav= pmass(1)/(4.0d0/3.0d0 * pi *pradius(1)**3)

  do i=1,Num_Spacetime
     call StarMapper_map(srho(i),sr(i),prho,pradius,zones)
!     srho(i) = rhoav
  enddo

  smass(1)=0
  smass(2)=4.0d0/3.0d0*pi*sr(2)**3 * srho(1)

  do i=3,szones
     dr = sr(i) - sr(i-1)
     smass(i) = smass(i-1) + 4*pi*sr(i-1)**2*srho(i-1)*dr
  enddo

!  do i=Num_Spacetime+1,szones
!     smass(i) = pmass(i-Num_Spacetime+maxi-1)
!  enddo


!  open(666,file="sdata2.dat",status="unknown",action="write")
!  do i=1,szones
!     write(666,"(I6,1P10E18.6)") i,sr(i),srho(i), & 
!         smass(i)
!  enddo

  snewtpot(1) = 0.0d0
  snewtpot(2) = smass(2)/sr(2)
  do i=3,szones
     dr = sr(i+1)-sr(i)
     snewtpot(i) = snewtpot(i-1) + smass(i-1)/sr(i-1)**2 * dr
  enddo

  difference = snewtpot(szones) - &
       smass(szones)/sr(szones)
  
  snewtpot = snewtpot - difference
     
#if 0  
  open(666,file="sdata.dat",status="unknown",action="write")
  do i=1,Num_Spacetime
     write(666,"(I6,1P10E18.6)") i,sr(i),srho(i), & 
         smass(i), snewtpot(i),0.0d0, &
         0.0d0,0.0d0
  enddo

  do i=Num_Spacetime+1,szones
!     write(*,"(I6,1P10E18.6)") i,srho(i),sr(i)
     if(sr(i).gt.5000) goto 20
     write(666,"(I6,1P10E18.6)") i,sr(i),srho(i), & 
         smass(i), snewtpot(i),pmass(i-Num_Spacetime+maxi), &
         prho(i-Num_Spacetime),pradius(i-Num_Spacetime+maxi)
  enddo
  close(666)
20 continue

#endif
!  stop

#if 0

! setup radii

  sr(1) = 0.0d0
  do i=2,Num_Spacetime
     sr(i) = sr(i-1) + dr
  enddo

! now get the consistent density  
  do i=1,Num_Spacetime
      call StarMapper_map(srho(i),sr(i),prho,pradius,zones)
  enddo


  smass(1) = 0.0d0
  do i=2,Num_Spacetime+Num_Radial
!     snewtpot(i) = snewtpot(i-1) + 
     dr = sr(i) - sr(i-1)
     smass(i) = smass(i-1) + 4*pi*sr(i)**2*srho(i)*dr
  enddo

  snewtpot(Num_Spacetime) = -ggrav*smass(Num_Spacetime)/sr(Num_Spacetime)

  do i=Num_Spacetime+Num_Radial-1,2,-1
     snewtpot(i) = snewtpot(i+1) - ggrav*smass(i)/sr(i)**2 * dr
  enddo
  snewtpot(1) = snewtpot(2)

  open(666,file="sdata.dat",status="unknown",action="write")
  do i=1,Num_Spacetime+Num_Radial
     write(666,"(I6,1P10E18.6)") i,sr(i), srho(i), smass(i), snewtpot(i)
  enddo
  close(666)

#endif


! note that the below is only (at best) first order accurate...
!  ppot(zones) = - ggrav*pmass(zones)/pradius(zones)

!  do i=zones-1,1,-1
!     dr = pradius(i+1) - pradius(i)
!     ppot(i) = ppot(i+1) - ggrav*pmass(i)/pradius(i)**2 * dr
!  enddo

!  write(*,*) pmass(1),pradius(1),dr

! debug output:
! open(666,file="testpot.dat",status="unknown",action="write")
! do i=1,zones
!    write(666,"(I6,1P10E18.6)") i,pradius(i),ppot(i)
! enddo
! close(666)
!!!

end subroutine CCCC_StarMapper_NewtPot
