#include "cctk.h"
#include "cctk_Parameters.h"
#include "cctk_Arguments.h"
#include "cctk_Functions.h"

subroutine CCCC_StarMapper_ReadProfile(CCTK_ARGUMENTS)


  implicit none

  DECLARE_CCTK_ARGUMENTS
  DECLARE_CCTK_PARAMETERS
  DECLARE_CCTK_FUNCTIONS

  character(len=512) filename
  integer filenamelength
  integer zones,n,i
 
  filenamelength = 512

  n = Num_Radial+Num_Spacetime
  zones = Num_Radial

  pradius(:) = 0.0d0
  pradius_center(:) = 0.0d0
  pmass(:) = 0.0d0  
  prho(:) = 0.0d0
  ppress(:) = 0.0d0
  ptemp(:) = 0.0d0
  peps(:) = 0.0d0  
  pvel(:) = 0.0d0
  pye(:) = 0.0d0
  pomega(:) = 0.0d0  
!  ppot(:) = 0.0d0



! Let's figure out the profile filename
  call CCTK_FortranString(filenamelength,Profile_File,filename)


  if (CCTK_EQUALS(Profile_Type,"SESAME")) then
     call ReadProfileSESAME(pradius,pmass,prho,&
          ppress,ptemp,peps,pvel,pye,pomega,zones, n, filename)
  else if (CCTK_EQUALS(Profile_Type,"PolyCott")) then
     call ReadProfilePolyCott(pradius,pmass,prho,&
          ppress,ptemp,peps,pvel,pye,pomega,zones, n, filename)
  else if (CCTK_EQUALS(Profile_Type,"HegerShort")) then
     call ReadProfileHegerShort(pradius,pmass,prho,&
          ppress,ptemp,peps,pvel,pye,pomega,zones, n, filename)
  else
     call CCTK_WARN(0,"Profile Type not implemented!")
  end if





! next:
! eos routines
! mapping routine
! gravity routine
! routine at paramcheck that looks for file

!  call CCTK_WARN(0,"STOPSTOPSTOP")

end subroutine CCCC_StarMapper_ReadProfile

subroutine ReadProfileSESAME(pradius,pmass,prho,&
     ppress,ptemp,peps,pvel,pye,pomega,zones, n, filename)

  use StarMapper_Module
  implicit none
  
! Format of SESAME initial profiles:
! radius,density,velocity,pressure,temperature,eps,?,mass,ye,?,zone #
! all units cgs, except temperature: MeV
!
!
  integer zones
  integer n
  CCTK_REAL pradius(n),pmass(n),prho(n),ppress(n),ptemp(n)
  CCTK_REAL peps(n),pvel(n),pye(n),pomega(n)


  CCTK_REAL buffer,pi
  character(len=512) filename

  integer i,ibuffer

  pi=3.14159265358979d0

  call CCTK_INFO("Reading Progenitor Model Profile")
  call CCTK_INFO(filename)
  
  open(666,file=trim(filename),status='unknown',form='formatted',action='read')
  do i=1,zones
     read(666,*) pradius(i),prho(i),pvel(i),ppress(i),ptemp(i), &
          peps(i),buffer,pmass(i),pye(i)
  enddo

  close(666)

! fudge innermost zone mass

  pmass(1) = 4.0d0/3.0d0 * pi * pradius(1)**3 * prho(1)
  peps(1:zones) = peps(1:zones)*mev_to_erg


end subroutine ReadProfileSESAME

subroutine ReadProfilePolyCott(pradius,pmass,prho,&
     ppress,ptemp,peps,pvel,pye,pomega,zones,n,filename)

  implicit none
  
! Format of PolyCott initial profiles:
! zone number, mass, radius, density, pressure, eps 
! all units cgs

  integer zones,n
  CCTK_REAL pradius(n),pmass(n),prho(n),ppress(n),ptemp(n)
  CCTK_REAL peps(n),pvel(n),pye(n),pomega(n)


  CCTK_REAL buffer,pi
  character(len=512) filename

  integer i,ibuffer

  pi=3.14159265358979d0

  call CCTK_INFO("Reading Progenitor Model Profile")
  call CCTK_INFO(filename)
  
  open(666,file=trim(filename),status='unknown',form='formatted',action='read')
  do i=1,zones
     read(666,*) ibuffer,pmass(i),pradius(i),prho(i),ppress(i),peps(i)
  enddo
  close(666)

  ptemp(:) = 0.0d0
  pomega(:) = 0.0d0
  pye(:) = 0.5d0
  pvel(:) = 0.0d0

end subroutine ReadProfilePolyCott

subroutine ReadProfileHegerShort(pradius,pmass,prho,&
     ppress,ptemp,peps,pvel,pye,pomega,zones,n,filename)

  use StarMapper_Module
  implicit none
  
! Format of HegerShort initial profile

! zone, mass, radius, temperature, density, velocity, ye, omega
! radius,density,velocity,pressure,temperature,eps,?,mass,ye,?,zone #
! all units cgs, except temperature: MeV
!
!
  integer zones,n
  CCTK_REAL pradius(n),pmass(n),prho(n),ppress(n),ptemp(n)
  CCTK_REAL peps(n),pvel(n),pye(n),pomega(n)

  CCTK_REAL buffer
  character(len=512) filename

  integer i,ibuffer

!  pi=3.14159265358979d0

  call CCTK_INFO("Reading Progenitor Model Profile: HegerShort type")
  call CCTK_INFO(filename)
  
  open(666,file=trim(filename),status='unknown',form='formatted',action='read')
  ! read number of lines in profile
  read(666,*) ibuffer
  do i=1,zones
     read(666,*) ibuffer,pmass(i),pradius(i),ptemp(i),prho(i),pvel(i),pye(i),pomega(i)
  enddo

  close(666)

! fudge innermost zone mass

  ptemp(1:zones) = ptemp(1:zones) * kboltz

!  pmass(1) = 4.0d0/3.0d0 * pi * pradius(1)**3 * prho(1)
!  peps(1:zones) = peps(1:zones)*mev_to_erg


end subroutine ReadProfileHegerShort
