#include "cctk.h"
#include "cctk_Parameters.h"
#include "cctk_Arguments.h"
#include "cctk_Functions.h"

subroutine CCCC_StarMapper_SetPressEps(CCTK_ARGUMENTS)

  use StarMapper_Module

  implicit none

  CCTK_INT ierr
  integer i,j,k,l,nx,ny,nz

  DECLARE_CCTK_ARGUMENTS
  DECLARE_CCTK_PARAMETERS
  DECLARE_CCTK_FUNCTIONS

! begin EOS Omni vars
  integer :: n,keytemp,anyerr,keyerr(1)
  real*8  :: xdummy(1)
! end EOS Omni vars

  nx = cctk_lsh(1)
  ny = cctk_lsh(2)
  nz = cctk_lsh(3)


  call CCTK_INFO("Setting consistent Pressures and EPSs")

!  write(*,"(1P10E14.5)") eps(10,5,5)/specific_internal_energy_factor
!  write(*,"(1P10E14.5)") press(10,5,5)/press_factor

  if(CCTK_EQUALS(Starmapper_EoS,"HotNuclear")) then
     ierr = EOS_SetGFs(cctkGH, EOS_StarMapperCall)
  else if(CCTK_EQUALS(Starmapper_EoS,"Hybrid")) then

     do i=1,nx
        do j=1,ny
           do k=1,nz
              ! set up pressure
              press(i,j,k) = press_factor * eosk * (rho(i,j,k)/density_factor)**gamma1
              eps(i,j,k) = press(i,j,k)/rho(i,j,k)/(gamma1-1.0d0)
!              cs2(i,j,k) = gamma1 * press(i,j,k)/rho(i,j,k) /  &
!                   (1.0d0 + eps(i,j,k) + press(i,j,k)/rho(i,j,k))
           enddo
        enddo
     enddo

  else if(CCTK_EQUALS(Starmapper_EOS,"Omni")) then
      n = 1
      keytemp = 1
      keyerr = 0
      anyerr = 0
      do i=1,nx
         do j=1,ny
            do k=1,nz
               if(Starmapper_eoskey.eq.4) then
                  call EOS_Omni_short(Starmapper_eoskey,keytemp,Starmapper_eos_rf_prec,&
                       n,rho(i,j,k),eps(i,j,k),temperature(i,j,k),y_e(i,j,k),&
                       xdummy,entropy(i,j,k),xdummy,xdummy,&
                       xdummy,xdummy,xdummy,keyerr,anyerr)
               else
                  call EOS_Omni_press(Starmapper_eoskey,keytemp,Starmapper_eos_rf_prec,n,&
                       rho(i,j,k),eps(i,j,k),xdummy,xdummy,press(i,j,k),keyerr,anyerr)
               endif
               if(anyerr.ne.0) then
                  call CCTK_WARN(0,"EOS problem in Starmapper!!!")
               endif
            enddo
         enddo
      enddo
   endif

!  write(6,"(1P10E16.8)") rho(4,4,4),eps(4,4,4),temperature(4,4,4),y_e(4,4,4)
!  stop "eh"
!  write(6,"(1P10E14.5)") press(10,5,5)/press_factor
!  write(6,"(1P10E14.5)") temperature(10,5,5)
!  write(6,"(1P10E14.5)") y_e(10,5,5)
!  call flush(6)


end subroutine CCCC_StarMapper_SetPressEps
