#include "cctk.h"
#include "cctk_Parameters.h"
#include "cctk_Arguments.h"
#include "cctk_Functions.h"

#include "util_ErrorCodes.h"
#include "util_Table.h"



subroutine CCCC_StarMapper_Setup(CCTK_ARGUMENTS)
  
  use StarMapper_Module

  implicit none
  DECLARE_CCTK_PARAMETERS
  DECLARE_CCTK_FUNCTIONS
  DECLARE_CCTK_ARGUMENTS


  CCTK_INT :: ierr, flags, table_handle, var
  CCTK_INT, dimension(3) :: indep_gfs
  CCTK_INT, dimension(2) :: dep_gfs
  CCTK_INT :: istat

  character(len=100) warnmessage


! Set up EOS calls; In the moment this is 
! hard-coded to only use EOS_GeneralHotNuclear
  
  flags = UTIL_TABLE_FLAGS_CASE_INSENSITIVE

  if (CCTK_EQUALS(Starmapper_EoS,"HotNuclear")) then

     call Util_TableCreate(table_handle, flags)
    
     call Util_TableSetString(ierr, table_handle, &
          "HotNuclear T", "EOS Name")

     call Util_TableSetString(ierr, table_handle, &
          "Rho Temperature ElectronFraction", &
          "Independent variable names")
  
     call Util_TableSetString(ierr, table_handle, &
          "Pressure SpecificInternalEnergy", &
          "Dependent variable names")

     call Util_TableSetInt(ierr, table_handle, &
          3, "N independent variables")

     call Util_TableSetInt(ierr, table_handle, &
          2, "N dependent variables")

     call CCTK_VarIndex(indep_gfs(1), "hydrobase::rho")
     call CCTK_VarIndex(indep_gfs(2), "EOS_GeneralHotNuclear::eos_temp")
     call CCTK_VarIndex(indep_gfs(3), "GRHydro::tracer[0]")

!     write(warnmessage,"(a20,i10)") "tracer index: ",indep_gfs(3)
!     call CCTK_WARN(1,warnmessage)

     call CCTK_VarIndex(dep_gfs(1), "hydrobase::press")
     call CCTK_VarIndex(dep_gfs(2), "hydrobase::eps")

     call Util_TableSetIntArray(ierr, table_handle, 3, &
           indep_gfs, "Independent GFs")

     call Util_TableSetIntArray(ierr, table_handle, 2, &
           dep_gfs, "Dependent GFs")
     
     EOS_StarMapperCall = EOS_SetupCall(table_handle)

  else if(CCTK_EQUALS(Starmapper_EoS,"Hybrid")) then
     gamma1 = Hybrid_Gamma
     eosk = Hybrid_K
  else if(CCTK_EQUALS(Starmapper_EOS,"Omni")) then
     if(CCTK_IsThornActive("EOS_Omni").eq.0) then
        call CCTK_WARN(0,"OUCH! EOS_Omni not active! :-(")
     endif
     ! check that we have storage for temperature and Y_e, just to be sure
     if(Starmapper_eoskey.eq.4) then

        istat = 0
        call CCTK_QueryGroupStorage(istat,cctkGH,"hydrobase::temperature")
        if(istat.eq.0) then
           call CCTK_WARN(0,"OUCH! Please enable storage for temperature in HydroBase")
        endif
        istat = 0
        call CCTK_QueryGroupStorage(istat,cctkGH,"hydrobase::Y_e")
        if(istat.eq.0) then
           call CCTK_WARN(0,"OUCH! Please enable storage for Y_e in HydroBase")
        endif
     endif

  else
     call CCTK_WARN(0,"OUCH! EoS selection not implemented in CCCC_Starmapper! :-(")
  endif

  call CCTK_INFO("Registered EOS call")

end subroutine CCCC_StarMapper_Setup
