#include "cctk.h"
#include "cctk_Parameters.h"
#include "cctk_Arguments.h"
#include "cctk_Functions.h"

subroutine CCCC_StarMapper_SetupMetric(CCTK_ARGUMENTS)

  use StarMapper_Module
  implicit none

  DECLARE_CCTK_ARGUMENTS
  DECLARE_CCTK_PARAMETERS
  DECLARE_CCTK_FUNCTIONS

  integer zones,i,czones
  character(len=512) filename
  integer filenamelength
  CCTK_REAL, allocatable :: cr(:), calp(:), cphi(:), cphi4(:)

  filenamelength=512
  czones = Num_Radial_CoCoNuT

  allocate(cr(czones))
  allocate(calp(czones))
  allocate(cphi(czones))
  allocate(cphi4(czones))

  cr(:) = 0.0d0
  calp(:) = 0.0d0
  cphi(:) = 0.0d0
  cphi4(:) = 0.0d0


  call CCTK_INFO("Setting up Profile Metric")

  if (use_coconut_spacetime_profile .eq. 1) then
    ! Read in CoCoNuT profile
    ! Get filename
    call CCTK_FortranString(filenamelength, CoCoNut_Profile_File, filename)
    call ReadProfileCoCoNuT(cr, calp, cphi, czones, filename)
    cphi4(:) = cphi(:) * cphi(:) * cphi(:) * cphi(:)
    ! Interpolate 1D arrays to salp and smetric. Set to 1 if extrapolation is
    ! is out of "physical" bounds
    do i=1,szones
      call StarMapper_map(salp(i),sr(i),calp,cr,czones)
      if (salp(i) .gt. 1.0d0) then
        salp(i) = 1.0d0
      end if
      call StarMapper_map(smetric(i),sr(i),cphi4,cr,czones)
      if (smetric(i) .lt. 1.0d0) then
        smetric(i) = 1.0d0
      end if
    enddo

  else
    ! Equation (8.50) of B. Schutz's 'A first course in numerical relativity'
    salp = sqrt((1.0d0 + 2.0d0 * snewtpot))
    smetric = (1.0d0 - 2.0d0 * snewtpot)
  end if


end subroutine CCCC_StarMapper_SetupMetric

! Read in initial spacetime CoCoNuT data
subroutine ReadProfileCoCoNuT(cr, calp, cphi, czones, filename)
  use StarMapper_Module
  implicit none

  integer czones, i
  CCTK_REAL cr(czones), calp(czones), cphi(czones)
  character(len=512) filename

  call CCTK_INFO("Reading in CoCoNuT Spacetime")
  call CCTK_INFO(filename)

  open(666,file=trim(filename),status='unknown',form='formatted',action='read')
  do i=1,czones
    read(666,*) cr(i), calp(i), cphi(i)
    cr(i) = cr(i) * 6.77269222552442d-06 ! Convert from cm to code units
  enddo
  close(666)

  ! Add any necessary unit conversions here
end subroutine ReadProfileCoCoNuT

